#!/usr/bin/env python
# coding: utf-8

"""
Create sentinel and singleton objects.

Copyright 2014, 2020 © Eddie Antonio Santos. MIT licensed.

With contributions from:
 - Simeon Visser (https://github.com/svisser)
 - WildCard65 (https://github.com/WildCard65)

This software is released under the MIT License.
https://opensource.org/licenses/MIT
"""

import inspect
from typing import Any, Dict, Optional, Tuple

__all__ = ["create"]
__version__ = "0.2.0"

_sinstances: Dict[type, object] = {}

_DEFAULT_MRO = (object,)


def _get_caller_module() -> Optional[str]:
    stack = inspect.stack()
    assert len(stack) > 1

    caller = stack[2][0]
    return caller.f_globals.get("__name__")


def _sentinel_failing_new(cls: type, *args: Tuple[Any, ...], **kwargs: Dict[str, Any]):
    raise TypeError("This sentinel object can not be allocated more than once")


def create(
    name: str = None,
    mro: Tuple[type, ...] = _DEFAULT_MRO,
    cls_dict: Dict[str, Any] = None,
    *args: Tuple[Any, ...],
    **kwargs: Dict[str, Any]
) -> object:
    """
    create(name, mro=(object,), cls_dict=None, ...) -> Sentinel instance

    Creates a new sentinel instance. This is a singleton instance kind
    of like the builtin None, and Ellipsis.

    Method resolution order (MRO) for the anonymous class can be
    specified (i.e., it can be a subclass). Provide the mro as tuple of
    all classes that it inherits from. If only one class, provide a
    1-tuple: e.g., (Cls,).

    Additionally extra class attributes, such as methods can be provided
    in the cls_dict dict. The following methods are provided, but
    can be overridden:

        __repr__()
            Returns the class name, similar to None and Ellipsis.
        __copy__()
        __deepcopy__()
            Always return the same singleton instance such that
            ``copy(Sentinel) is Sentinel`` is true.
        __reduce__()
            Provided for proper pickling prowess. That is,
            ``pickle.loads(pickle.dumps(Sentinel)) is Sentinel`` is
            true.

    Finally, the remain arguments and keyword arguments are passed to
    the super class's __init__().  This is helpful when for
    instantiating base classes such as a tuple.
    """

    if name is None:
        try:
            from varname import varname  # type: ignore
        except ImportError as error:
            raise ImportError(
                "Cannot infer variable name without varname library; "
                "please install the varname PyPI package to enable this feature: "
                "pip install sentinel[varname]"
            ) from error
        name = varname()

    _cls_dict = {
        # make the singleton always belong to the module of its caller.
        # If we don't do this, pickling using __reduce__() will fail!
        "__module__": _get_caller_module(),
        # Provide a nice, clean, self-documenting __repr__
        "__repr__": lambda _: name,
        # Provide a hook for pickling the sentinel.
        "__reduce__": lambda _: name,
        # Provide a copy and deepcopy implementation which simply return the
        # same exact instance.
        "__copy__": lambda self: self,
        "__deepcopy__": lambda self, _: self,
    }

    if mro == _DEFAULT_MRO:
        # Add an empty slots tuple if we're only inheritting from 'object'
        _cls_dict["__slots__"] = ()
    elif object in mro:
        # Validate if object is the last type in the provided mro list.
        assert mro[-1] is object, "object should ALWAYS the last type in a mro"

    class _SentinelMeta(type(mro[0])):  # Inherit from the base type of the first type.
        instance = property(
            fget=_sinstances.get, doc="""Gets the instance of this sentinel type."""
        )

        def __call__(self, *args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> object:
            try:
                return _sinstances[self]
            except KeyError:
                inst = super().__call__(
                    *args, **kwargs
                )  # We use super() incase __call__ was overriden.
                _sinstances[self] = inst

                return inst

    if cls_dict is not None:
        _cls_dict.update(cls_dict)
    _sentinel = _SentinelMeta("_Sentinel", mro, _cls_dict)

    inst = _sentinel(*args, **kwargs)
    _sentinel.__new__ = _sentinel_failing_new

    return inst



# -*- coding: utf-8 -*-
from setuptools import setup

modules = \
['sentinel']
extras_require = \
{'varname': ['varname>=0.1']}

setup_kwargs = {
    'name': 'sentinel',
    'version': '0.3.0',
    'description': 'Create sentinel objects, akin to None, NotImplemented, Ellipsis',
    'long_description': '*************************************************\nsentinel — create sentinel and singleton objects\n*************************************************\n\n|Tests| |PyPI version|\n\n.. |Tests| image:: https://github.com/eddieantonio/sentinel/workflows/Test%20and%20Lint/badge.svg\n   :target: https://github.com/eddieantonio/sentinel/actions?query=workflow%3A%22Test+and+Lint%22\n.. |PyPI version| image:: https://img.shields.io/pypi/v/sentinel\n   :target: https://pypi.org/project/sentinel/\n\nCreates simple sentinel objects.\n\n\nInstall\n=======\n\nBasic features::\n\n   pip install sentinel\n\nwith extra magic features powered by python-varname_::\n\n   pip install \'sentinel[varname]\'\n\n\nWhat is a sentinel?\n===================\n\nSentinels_ are singleton_ objects that typically represent some\nterminating (end) condition or have a special, symbolic meaning. Python\'s built-in\n``None`` is a sentinel. Python also has other sentinels like ``NotImplemented`` and\n``Ellipsis``.\n\nIf you want to create your own sentinels, use this library! Make your calls to\n``dict.get()`` more meaningful! You can replace the ``object()`` idiom with a sentinel:\n\n.. code-block:: python\n\n   d = {"a": 1, "b": None}\n\n   # Before sentinel:\n   missing = object()\n   if d.get("c", missing) is missing:\n       ... # do some stuff\n\n   # After sentinel:\n   Missing = sentinel.create()\n   if d.get("c", Missing) is Missing:\n       ... # do some stuff\n\n\nFeatures\n--------\n\n - sentinels are unique\n - sentinels are singletons — the **only** instance of their own anonymous class\n - sentinels can be used with ``is`` comparisons\n - sentinels can be used with ``pickle``\n - sentinels can be used with ``copy.deepcopy``\n - you can **add** arbitrary attributes and methods to sentinels\n - sentinels have a nice, self-documenting ``__repr__``!\n\nUsage\n=====\n\nCreate a sentinel:\n\n>>> import sentinel\n>>> MySentinel = sentinel.create("MySentinel")\n>>> Sentinel\nSentinel\n\nIf you have python-varname_ installed, or installed this module using\n``pip install \'sentinel[varname]\'``, ``sentinel.create()`` can infer the name\nfrom the assignment expression:\n\n.. code-block:: python\n\n   import sentinel\n\n   MySentinel = sentinel.create()\n\n   print(MySentinel)  # prints `MySentinel`\n\n\nExample\n-------\n\nSentinels are useful when other objects such as ``None``, ``False``,\n``0``, ``-1``, are valid values within some data structure. For example, setting\ndefault values when all other values are valid with:\n``dict.setdefault()``:\n\n.. code-block:: python\n\n   d = {"stdout": None, "stdin": 0, "EOF": -1}\n\n   MissingEntry = sentinel.create()\n\n   [d.setdefault(key, MissingEntry) for key in ("stdin", "stdout", "stderr")]\n   [0, None, MissingEntry]\n\nAlternatively, using ``dict.get()`` when fetching values:\n\n>>> d = {"stdout": None, "stdin": 0, "EOF": -1}\n>>> d.get("stdout", MissingEntry)\nNone\n>>> d.get("stdin", MissingEntry)\n0\n>>> d.get("stderr", MissingEntry)\nMissingEntry\n\nSince a new sentinel can never occur in the original dictionary, you can tell which\nentries are missing or unset in a dictionary in a self-documenting way:\n\n.. code-block:: python\n\n   Unset = sentinel.create()\n   if d.get("stdin", Unset) is Unset:\n       stdin = 0  # some reasonable default\n\n\nAdding extra methods and class attributes\n-----------------------------------------\n\nSentinels may also inherit from base classes, or implement extra methods.\n\nConsider a binary search tree with two kinds of nodes: interior nodes\n(``Node``) which contain some payload and leaves (``Leaf``), which simply\nterminate traversal.\n\nTo create singleton leaf which implements a ``search`` method and an\n``is_leaf`` property, you may provide any extra class attributes in the\n``cls_dict`` keyword argument. The following is a full example of both\nthe singleton ``Leaf`` and its ``Node`` counterpart:\n\n.. code-block:: python\n\n    def _search_leaf(self, key):\n        raise KeyError(key)\n\n    Leaf = sentinel.create(\'Leaf\', cls_dict={\n        \'search\': _search_leaf,\n        \'is_leaf\': property(lambda self: True)\n    })\n\n    class Node(object):\n        def __init__(self, key, payload, left=Leaf, right=Leaf):\n            self.left = left\n            self.right = right\n            self.key = key\n            self.payload = payload\n\n        def search(self, key):\n            if key < self.key:\n                return self.left.search(key)\n            elif key > self.key:\n                return self.right.search(key)\n            else:\n                return self.payload\n\n        is_leaf = property(lambda: false)\n\nExample usage:\n\n>>> tree = Node(2, \'bar\', Node(1, \'foo\'), Node(3, \'baz\'))\n>>> tree.search(1)\n\'foo\'\n>>> tree.search(4)\nTraceback (most recent call last):\n    ...\nKeyError: 2\n\n\nContributing\n============\n\nThis project uses Poetry_. To contribute to the codebase, make sure to `install poetry`_,\nWith Poetry installed, clone then repo, then within the repo directory, install the developer dependencies::\n\n    $ poetry install --extras varname\n\nNext, I recommend you do all development tasks within the ``poetry shell``::\n\n    $ poetry shell\n    (sentinel-nUnrocCf-py3.9) $ black .\n    (sentinel-nUnrocCf-py3.9) $ pytest\n\n.. _Sentinels: http://en.wikipedia.org/wiki/Sentinel_nodes\n.. _singleton: http://en.wikipedia.org/wiki/Singleton_pattern\n.. _Poetry: https://python-poetry.org/\n.. _install poetry: https://python-poetry.org/docs/#installation\n.. _python-varname: https://github.com/pwwang/python-varname\n',
    'author': 'Eddie Antonio Santos',
    'author_email': 'easantos@ualberta.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/eddieantonio/sentinel',
    'py_modules': modules,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
